#include <v8.h>
#include <node.h>
#include <node_buffer.h>

#include "ft2build.h"
#include FT_FREETYPE_H
#include FT_OUTLINE_H

#include "v8_helpers.h"

#include "util.h"
#include "freetype.h"

using namespace node;
using namespace v8;

void init(Handle<Object> target) {
  HandleScope scope;

  NODE_SET_METHOD(target, "initFreeType" , freetype::InitFreeType);
  NODE_SET_METHOD(target, "doneFreeType" , freetype::DoneFreeType);
  NODE_SET_METHOD(target, "newMemoryFace", freetype::NewMemoryFace);
  NODE_SET_METHOD(target, "doneFace"     , freetype::DoneFace);
  NODE_SET_METHOD(target, "setCharSize"  , freetype::SetCharSize);
  NODE_SET_METHOD(target, "getCharIndex" , freetype::GetCharIndex);
  NODE_SET_METHOD(target, "loadGlyph"    , freetype::LoadGlyph);
}
NODE_MODULE(freetype, init)

/* FreeType */


V8_METHOD(freetype::InitFreeType) {
  HandleScope scope;

  FT_Library *library = (FT_Library*) malloc(sizeof(FT_Library));
  V8::AdjustAmountOfExternalAllocatedMemory(sizeof(FT_Library));

  if (FT_Init_FreeType(library)) {
    V8_THROW(Exception::TypeError(String::New("Error initializing freetype.")));
  }

  Handle<Object> libraryObj = Object::New();
  libraryObj->Set(String::NewSymbol("library"), External::New(library));
  V8_RETURN(scope.Close(libraryObj));
}

V8_METHOD(freetype::DoneFreeType) {
  HandleScope scope;

  Handle<Object> libraryObj = args[0]->ToObject();
  Handle<String> librarySymbol = String::NewSymbol("library");
  FT_Library* library = static_cast<FT_Library*>(External::Cast(*libraryObj->Get(librarySymbol))->Value());

  FT_Error error = FT_Done_FreeType(*library);
  free(library);
  V8::AdjustAmountOfExternalAllocatedMemory(-sizeof(FT_Library));

  if (error) {
    V8_THROW(Exception::TypeError(String::New("Error unloading face.")));
  }

  V8_RETURN(Undefined());
}


/* Face */


V8_METHOD(freetype::NewMemoryFace) {
  HandleScope scope;

  Handle<Object> libraryObj = args[0]->ToObject();
  Handle<String> librarySymbol = String::NewSymbol("library");
  FT_Library* library = static_cast<FT_Library*>(External::Cast(*libraryObj->Get(librarySymbol))->Value());

  if (!Buffer::HasInstance(args[1])) {
    V8_THROW(Exception::Error(String::New("Second argument must be a buffer")));
  }

  Local<Object> bufferObj = args[1]->ToObject();
  char *bufferData = Buffer::Data(bufferObj);
  size_t bufferLength = Buffer::Length(bufferObj);

  FT_Face *face = (FT_Face*) malloc(sizeof(FT_Face));
  V8::AdjustAmountOfExternalAllocatedMemory(sizeof(FT_Face));

  FT_Byte* file_base  = (FT_Byte*) bufferData;
  FT_Long  file_size  = (FT_Long) bufferLength;
  FT_Long  face_index = (FT_Long) (FT_Long) args[2]->Int32Value();

  // // TODO: One or more of:
  // //         * Enable/Disable this code with macros
  // //         * Implement similar functionality in test code.
  // printf("FT_New_Memory_Face(%p - %d, %p, %ld, %ld, %p)\n",
  //        *library, (int) *library,
  //        file_base,
  //        file_size,
  //        face_index,
  //        face);
  //
  // printf("  file_base(%2x, %2x, %2x, %2x, %2x)\n",
  //        file_base[0], file_base[1], file_base[2], file_base[3], file_base[4]);

  FT_Error error =
    FT_New_Memory_Face(*library,
                       file_base,
                       file_size,
                       face_index,
                       face);

  if (error) {
    free(face);
    V8::AdjustAmountOfExternalAllocatedMemory(-sizeof(FT_Face));
    V8_THROW(Exception::TypeError(String::New("Error loading face.")));
  }

  Handle<Object> faceObj = Object::New();
  faceObj->Set(String::NewSymbol("face"), External::New(face));
  faceObj->Set(String::NewSymbol("glyph"), Null());

  faceObj->Set(String::NewSymbol("num_glyphs"), Int32::New((*face)->num_glyphs));

  faceObj->Set(String::NewSymbol("family_name"), String::New((*face)->family_name));
  faceObj->Set(String::NewSymbol("style_name"), String::New((*face)->style_name));

  faceObj->Set(String::NewSymbol("units_per_EM"), Uint32::New((*face)->units_per_EM));
  faceObj->Set(String::NewSymbol("ascender"), Int32::New((*face)->ascender));
  faceObj->Set(String::NewSymbol("descender"), Int32::New((*face)->descender));
  faceObj->Set(String::NewSymbol("height"), Int32::New((*face)->height));

  V8_RETURN(scope.Close(faceObj));
}

V8_METHOD(freetype::DoneFace) {
  HandleScope scope;

  Handle<String> faceSymbol = String::NewSymbol("face");
  Handle<Object> faceObj = args[0]->ToObject();
  FT_Face* face = static_cast<FT_Face*>(External::Cast(*faceObj->Get(faceSymbol))->Value());

  FT_Error error = FT_Done_Face(*face);
  free(face);
  V8::AdjustAmountOfExternalAllocatedMemory(-sizeof(FT_Face));
  faceObj->Set(faceSymbol, Null());

  if (error) {
    V8_THROW(Exception::TypeError(String::New("Error unloading face.")));
  }

  V8_RETURN(Undefined());
}

V8_METHOD(freetype::SetCharSize) {
  HandleScope scope;

  Handle<String> faceSymbol = String::NewSymbol("face");
  Handle<Object> faceObj = args[0]->ToObject();
  FT_Face face = *static_cast<FT_Face*>(External::Cast(*faceObj->Get(faceSymbol))->Value());

  FT_F26Dot6  char_width      = args[1]->Uint32Value();
  FT_F26Dot6  char_height     = args[2]->Uint32Value();
  FT_UInt     horz_resolution = args[3]->Uint32Value();
  FT_UInt     vert_resolution = args[4]->Uint32Value();

  FT_Error error =
    FT_Set_Char_Size(face,
                     char_width, char_height,
                     horz_resolution, vert_resolution);

  if (error) {
    V8_THROW(Exception::TypeError(String::New("Error setting char size.")));
  }

  V8_RETURN(Undefined());
}

V8_METHOD(freetype::GetCharIndex) {
  HandleScope scope;

  Handle<String> faceSymbol = String::NewSymbol("face");
  Handle<Object> faceObj = args[0]->ToObject();
  FT_Face* face = static_cast<FT_Face*>(External::Cast(*faceObj->Get(faceSymbol))->Value());

  FT_ULong charcode = (FT_ULong) args[1]->Uint32Value();

  FT_UInt result = FT_Get_Char_Index(*face, charcode);

  V8_RETURN(scope.Close(Uint32::New(result)));
}

V8_METHOD(freetype::LoadGlyph) {
  HandleScope scope;

  Handle<String> faceSymbol = String::NewSymbol("face");
  Handle<Object> faceObj = args[0]->ToObject();
  FT_Face face = *static_cast<FT_Face*>(External::Cast(*faceObj->Get(faceSymbol))->Value());

  FT_UInt  glyph_index = (FT_UInt) args[1]->Uint32Value();
  FT_Int32 load_flags = (FT_Int32) args[2]->Int32Value();

  FT_Error error =
    FT_Load_Glyph(face, glyph_index, load_flags);

  if (error) {
    V8_THROW(Exception::TypeError(String::New("Error loading glyph.")));
  }

  Local<Object> glyphObj = Object::New();
  faceObj->Set(String::NewSymbol("glyph"), glyphObj);
  Local<Object> advanceObj = Object::New();
  glyphObj->Set(String::NewSymbol("advance"), advanceObj);
  advanceObj->Set(String::NewSymbol("x"), Int32::New(face->glyph->advance.x));
  advanceObj->Set(String::NewSymbol("y"), Int32::New(face->glyph->advance.y));
  Local<Object> outlineObj = Object::New();
  glyphObj->Set(String::NewSymbol("outline"), outlineObj);
  short nContours = face->glyph->outline.n_contours;
  short nPoints = face->glyph->outline.n_points;
  outlineObj->Set(String::NewSymbol("nContours"), Int32::New(nContours));
  outlineObj->Set(String::NewSymbol("nPoints"  ), Int32::New(nPoints));

  size_t size;
  size = nPoints * 2 * sizeof(signed long);
  Local<Value> points = newInt32Array(size);
  fillArray(points, face->glyph->outline.points, size);
  outlineObj->Set(String::NewSymbol("points"), points);

  size = nPoints * sizeof(char);
  Local<Value> tags = newInt8Array(size);
  fillArray(tags, face->glyph->outline.tags, size);
  outlineObj->Set(String::NewSymbol("tags"), tags);

  size = nContours * sizeof(short);
  Local<Value> contours = newInt16Array(size);
  fillArray(contours, face->glyph->outline.contours, size);
  outlineObj->Set(String::NewSymbol("contours"), contours);

  outlineObj->Set(String::NewSymbol("flags"  ), Int32::New(face->glyph->outline.flags));

  V8_RETURN(scope.Close(glyphObj));
}
